#include <antiglut.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#ifdef __WINDOWS__
#include <io.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif

#include "mvlib.h"

int printOglError(char *file, int line)
{
    /* Returns 1 if an OpenGL error occurred, 0 otherwise. */
    GLenum glErr;
    int    retCode = 0;

    glErr = glGetError();
    while (glErr != GL_NO_ERROR)
    {
        printf("glError in file %s @ line %d\n", file, line);
        retCode = 1;
        glErr = glGetError();
    }
    return retCode;
}

#define printOpenGLError() printOglError(__FILE__, __LINE__)

struct complete_shader
{
	GLchar *vert_src;
	GLchar *frag_src;
	GLuint vert;
	GLuint frag;
	GLuint prog;
};

/*
 * Get the location of a uniform variable
 */
static
GLint getUniLoc(GLuint program, const GLchar *name)
{
    GLint loc;
    loc = glGetUniformLocation(program, name);
    if (loc == -1)
        printf("No such uniform named \"%s\"\n", name);
    printOpenGLError();  /* Check for OpenGL errors */
    return loc;
}

/*
 * Print out the information log for a shader object
 */
static
void printShaderInfoLog(GLuint shader)
{
    int infologLength = 0;
    int charsWritten  = 0;
    GLchar *infoLog;
    printOpenGLError();  /* Check for OpenGL errors */
    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infologLength);
    printOpenGLError();  /* Check for OpenGL errors */
    if (infologLength > 0)
    {
        infoLog = (GLchar *)malloc(infologLength);
        if (infoLog == NULL)
        {
            printf("ERROR: Could not allocate InfoLog buffer\n");
            exit(1);
        }
        glGetShaderInfoLog(shader, infologLength, &charsWritten, infoLog);
        printf("Shader InfoLog:\n%s\n\n", infoLog);
        free(infoLog);
    }
    printOpenGLError();  /* Check for OpenGL errors */
}

/* 
 * Print out the information log for a program object
 */
static
void printProgramInfoLog(GLuint program)
{
    int infologLength = 0;
    int charsWritten  = 0;
    GLchar *infoLog;
    printOpenGLError();  /* Check for OpenGL errors */
    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infologLength);
    printOpenGLError();  /* Check for OpenGL errors */
    if (infologLength > 0)
    {
        infoLog = (GLchar *)malloc(infologLength);
        if (infoLog == NULL)
        {
            printf("ERROR: Could not allocate InfoLog buffer\n");
            exit(1);
        }
        glGetProgramInfoLog(program, infologLength, &charsWritten, infoLog);
        printf("Program InfoLog:\n%s\n\n", infoLog);
        free(infoLog);
    }
    printOpenGLError();  /* Check for OpenGL errors */
}

static char *
readGLprog(char *filename)
{
    FILE *fh;
    int shader_size;
    char name[256] = "glprogs/";
    int count;
    char *shaderText;
    strcat(name, filename);
    fh = fopen(name, "r");
    if (!fh)
	{
		perror("fopen");
        return NULL;
	}
    fseek(fh, 0, SEEK_END);
    shader_size = ftell(fh);
    fseek(fh, 0, SEEK_SET);
    if(!shader_size)
	{
		printf("Empty shader source \"%s\"\n", name);
		fclose(fh);
        return NULL;
	}
    shaderText = (GLchar *)malloc(shader_size + 1);
    count = (int)fread(shaderText, 1, shader_size, fh);
    if(count != shader_size)
    {
    	fclose(fh);
    	free(shaderText);
    	return NULL;
    }
    shaderText[count] = '\0';
    fclose(fh);
    return shaderText;
}

static void
freeShader(struct complete_shader *shader)
{
	if(shader->prog)
	{
		glDetachShader(shader->prog, shader->vert);
		glDetachShader(shader->prog, shader->frag);
		glDeleteProgram(shader->prog);
		shader->prog = 0;
	}
	if(shader->vert)
	{
		glDeleteShader(shader->vert);
		shader->vert = 0;
	}
	if(shader->frag)
	{
		glDeleteShader(shader->frag);
		shader->frag = 0;
	}
	if(shader->vert_src)
	{
		free(shader->vert_src);
		shader->vert_src = NULL;
	}
	if(shader->frag_src)
	{
		free(shader->frag_src);
		shader->frag_src = NULL;
	}
}

static int
readShader(char *vertname,char *fragname,
	struct complete_shader *shader)
{
    GLint  vertCompiled, fragCompiled;    /* status values */
    GLint  linked;
	shader->vert_src = shader->frag_src = NULL;
	shader->vert = shader->frag = shader->prog = 0;
	if(!(shader->vert_src = readGLprog(vertname)))
		return 0;
	if(!(shader->frag_src = readGLprog(fragname)))
	{
		freeShader(shader);
		return 0;
	}
    shader->vert = glCreateShader(GL_VERTEX_SHADER);
    shader->frag = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(shader->vert, 1, (const GLchar**)&shader->vert_src, NULL);
    glShaderSource(shader->frag, 1, (const GLchar**)&shader->frag_src, NULL);
    /*
     * Compile vertex shader and print out compiler log.
     */
    glCompileShader(shader->vert);
    printOpenGLError();
    glGetShaderiv(shader->vert, GL_COMPILE_STATUS, &vertCompiled);
    printShaderInfoLog(shader->vert);
    /*
     * Compile fragment shader and print out compiler log.
     */
    glCompileShader(shader->frag);
    printOpenGLError();
    glGetShaderiv(shader->frag, GL_COMPILE_STATUS, &fragCompiled);
    printShaderInfoLog(shader->frag);
    if (vertCompiled && fragCompiled)
    {
		shader->prog = glCreateProgram();
		glAttachShader(shader->prog, shader->vert);
		glAttachShader(shader->prog, shader->frag);
		/*
		 * Link the program and print out the info log
		 */
		glLinkProgram(shader->prog);
		printOpenGLError();
		glGetProgramiv(shader->prog, GL_LINK_STATUS, &linked);
		printProgramInfoLog(shader->prog);
		if(linked)
			return 1;
    }
    freeShader(shader);
    return 0;
}

/* ***********************************************************************
   ********************************************************************* */

struct complete_shader unitshader;
struct complete_shader perceptive;

static int
installShaders()
{
	readShader("unit.vert", "unit.frag", &unitshader);
	if(!unitshader.prog)
		return 0;
    glUseProgram(unitshader.prog);
    glUniform3f(getUniLoc(unitshader.prog, "LightPosition"), 0, 0, 4.0f);

	readShader("perc.vert", "white.frag", &perceptive);
	if(!perceptive.prog)
		return 0;
/*
	readShader("unit.vert", "unit.frag", &unitshader);
	if(!unitshader.prog || !perceptive.prog)
		return 0;
*/
    /* Set up initial uniform values */
/*
*/

    return 1;
}

static void
freeShaders()
{
	freeShader(&perceptive);
	freeShader(&unitshader);
}

/* ***********************************************************************
	drawing stuff
   ********************************************************************* */

static void drawCube(void)
{
    #define	size 1.0f
/*
    #define scale 0.2f
    #define	delta 0.1f
*/
    #define scale 1.0f
    #define	delta 0.0f

    float A[3] = { size,  size,  size * scale + delta };
    float B[3] = { size,  size, -size * scale + delta };
    float C[3] = { size, -size, -size * scale };
    float D[3] = { size, -size,  size * scale };
    float E[3] = {-size,  size,  size * scale + delta };
    float F[3] = {-size,  size, -size * scale + delta };
    float G[3] = {-size, -size, -size * scale };
    float H[3] = {-size, -size,  size * scale };

    float I[3] = { 1.0f,  0.0f,  0.0f};
    float K[3] = {-1.0f,  0.0f,  0.0f};
    float L[3] = { 0.0f,  0.0f, -1.0f};
    float M[3] = { 0.0f,  0.0f,  1.0f};
    float N[3] = { 0.0f,  1.0f,  0.0f};
    float O[3] = { 0.0f, -1.0f,  0.0f};


    glBegin(GL_QUADS);

    glNormal3fv(I);

    glTexCoord2f(1,1);
    glVertex3fv(D);
    glTexCoord2f(0,1);
    glVertex3fv(C);
    glTexCoord2f(0,0);
    glVertex3fv(B);
    glTexCoord2f(1,0);
    glVertex3fv(A);

    glNormal3fv(K);

    glTexCoord2f(1,1);
    glVertex3fv(G);
    glTexCoord2f(0,1);
    glVertex3fv(H);
    glTexCoord2f(0,0);
    glVertex3fv(E);
    glTexCoord2f(1,0);
    glVertex3fv(F);

    glNormal3fv(L);

    glTexCoord2f(1,1);
    glVertex3fv(C);
    glTexCoord2f(0,1);
    glVertex3fv(G);
    glTexCoord2f(0,0);
    glVertex3fv(F);
    glTexCoord2f(1,0);
    glVertex3fv(B);

    glNormal3fv(M);

    glTexCoord2f(1,1);
    glVertex3fv(H);
    glTexCoord2f(0,1);
    glVertex3fv(D);
    glTexCoord2f(0,0);
    glVertex3fv(A);
    glTexCoord2f(1,0);
    glVertex3fv(E);

    glNormal3fv(N);

    glTexCoord2f(1,1);
    glVertex3fv(E);
    glTexCoord2f(0,1);
    glVertex3fv(A);
    glTexCoord2f(0,0);
    glVertex3fv(B);
    glTexCoord2f(1,0);
    glVertex3fv(F);

    glNormal3fv(O);
    
    glTexCoord2f(1,1);
    glVertex3fv(G);
    glTexCoord2f(0,1);
    glVertex3fv(C);
    glTexCoord2f(0,0);
    glVertex3fv(D);
    glTexCoord2f(1,0);
    glVertex3fv(H);
    
    glEnd();
#undef size
#undef scale
#undef delta

}

static void NextClearColor(void)
{
    static int color = 0;

    switch( color++ )
    {
        case 0:  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
             break;
        case 1:  glClearColor(0.2f, 0.2f, 0.3f, 1.0f);
             break;
        default: glClearColor(0.7f, 0.7f, 0.7f, 1.0f);
             color = 0;
             break;
    }
}

/*
 * FIXME: port camera control from LibV!
 */
static GLfloat rot_x  = 0, rot_y = 0;	/* initial rotation angles */
static GLfloat cam_origin[3] = {0,0,0};	/* initial position of camera */

GLfloat scene_mins[3] = {-100, -100, 10};
GLfloat scene_maxs[3] = {100, 100, 2000};
GLfloat cube_scales_mins[3] = {1,1,10};
GLfloat cube_scales_maxs[3] = {20,20,50};

static void
init_frustum(int range_idx)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	switch(range_idx)
	{
		case 0:
			glFrustum(-1,1,-1,1,1,1000);
			break;
		case 1:
			glFrustum(-3,3,-3,3,9,3000);
			break;
		default:
			glFrustum(-10,10,-10,10,100,10000);
			break;
	}
	glMatrixMode(GL_MODELVIEW);
}

/*
 * returns GLfloat in range [min,max]
 */
static GLfloat
frand(GLfloat min,GLfloat max)
{
	GLfloat t = (GLfloat)rand() / (GLfloat)RAND_MAX;
	return min + t * (max - min);
}

/*
 * draws scaled and translated cube
 */
void
new_cube()
{
	int i;
	GLfloat origin[3];
	GLfloat scales[3];
	for(i=0;i<3;i++)
	{
		origin[i] = frand(scene_mins[i],scene_maxs[i]);
		scales[i] = frand(cube_scales_mins[i],cube_scales_maxs[i]);
	}
	glPushMatrix();
	glTranslatef(-origin[0],-origin[1],-origin[2]);
	glScalef(scales[0],scales[1],scales[2]);
	glColor3f(frand(0.3f,0.6f),frand(0.3f,0.6f),frand(0.3f,0.6f));
	drawCube();
	glPopMatrix();
}

#define	NUM_CUBES	100

GLuint	scene;

static void
create_scene()
{
	int i;
	scene = glGenLists(1);
	srand(2);
	glNewList(scene, GL_COMPILE);
	for(i = 0;i < NUM_CUBES;i++)
		new_cube();
	glEndList();
}

static void
free_scene()
{
	glDeleteLists(scene, 1);
}

int draw_initial = 1;
int draw_perceptive = 1;

static
void display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glTranslatef(cam_origin[0],cam_origin[1],cam_origin[2]);
	glRotatef(rot_x,1,0,0);
	glRotatef(rot_y,0,1,0);

	/*
	 * Initial
	 */
	if(draw_initial)
	{
		glUseProgram(unitshader.prog);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glCallList(scene);
	}
	/*
	 * Perceptive
	 */
	if(draw_perceptive)
	{
		glUseProgram(perceptive.prog);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonOffset(-1, -1);
		glCallList(scene);
		glDisable(GL_POLYGON_OFFSET_LINE);
	}
	glFlush();
	agSwap();
}

static int
handle_event_check_break(ui_event_p in)
{
	GLfloat ts;
	if(in->state & UI_CONTROL_MASK)
		ts = 10;
	else
		ts = 1;
	if(in->state & UI_ALT_MASK)
		ts *= 10;
	switch(in->type)
	{
		case ui_key_event:
			if(!in->key.press)
				break;
			switch(in->key.scan)
			{
				case sc_escape:
					return 1;
		        case sc_b:
		            NextClearColor();
		            break;
				case sc_wheelup:
				case sc_minus:
				case sc_kp_subtract:
					cam_origin[2] += ts;
					break;
				case sc_wheeldown:
				case sc_equal:
				case sc_kp_add:
					cam_origin[2] -= ts;
					break;
				case sc_space:
				case sc_kp_begin:
					rot_x = rot_y = 0;
					vec3Set(cam_origin,0,0,0);
					break;
				/* rot_y */
				case sc_left:
					rot_y -= 0.1f;
					break;
				case sc_right:
					rot_y += 0.1f;
					break;
				/* rot_x */
				case sc_up:
					rot_x -= 0.1f;
					break;
				case sc_down:
					rot_x += 0.1f;
					break;
				/* XY pan */
				case sc_kp_left:
					cam_origin[0] -= ts;
					break;
				case sc_kp_right:
					cam_origin[0] -= ts;
					break;
				case sc_kp_up:
					cam_origin[1] += ts;
					break;
				case sc_kp_down:
					cam_origin[1] -= ts;
					break;
				case sc_i:
					draw_initial = draw_initial ? 0 : 1;
					break;
				case sc_p:
					draw_perceptive = draw_perceptive ? 0 : 1;
					break;
				case sc_q:
					init_frustum(0);
					break;
				case sc_w:
					init_frustum(1);
					break;
				case sc_e:
					init_frustum(2);
					break;
			}
			break;
		case ui_pointer_motion_event:
			if(in->button & UI_LBUTTON_MASK) /* XY rot */
			{
				rot_y -= in->client_dx * ts * 0.1f;
				rot_x -= in->client_dy * ts * 0.1f;
			}
			if(in->button & UI_RBUTTON_MASK) /* XY pan */
			{
				cam_origin[0] += ts * in->client_dx;
				cam_origin[1] -= ts * in->client_dy;
			}
			break;
		case ui_wm_event:
			switch(in->configure.wm_event)
			{
				case ui_wm_close:
					return 1;
			}
			break;
		default:
			break;
	}
	return 0;
}

static void main_loop()
{
	ui_event_t ev;
	for(;;)
	{
		display();
		agSleep(10);
		while(uiPending())
		{
			uiNextEvent(&ev);
			if(handle_event_check_break(&ev))
				return;
		}
	}
}

/*
 * Main
 */

int main(void)
{
	agWindowConfig("[CTRL+ALT]+mouse to control. I,P - Toggle initial/perceptive",
		800, 800, GL_FALSE);
	if(!agInit(AG_REQUIRED_2_0,GL_TRUE,GL_FALSE,GL_FALSE))
		return -1;
	printf("GLSL version: %s\n",glGetString(GL_SHADING_LANGUAGE_VERSION));
	init_frustum(0);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	NextClearColor();
	if(!installShaders())
	{
		agShutdown();
		return -1;
	}
	create_scene();
	main_loop();
	free_scene();
	freeShaders();
	agShutdown();
	return 0;
}
